# dbl_pend.py
#
# Sample Webots controller file for driving a
# fully actuated double pendulum.  With both actuators
# set to zero torque the device will act like a chaotic
# double pendulum. Otherwise this code implements
# PD position control to drive towards a pose.
#
# No copyright, 2020-2022, Garth Zeglin.  This file is
# explicitly placed in the public domain.

print("loading pancake.py...")

# Import the Webots simulator API.
from controller import Robot
from controller import Keyboard

# Import the standard Python math library.
import math

# Define the time step in milliseconds between
# controller updates.
EVENT_LOOP_DT = 20

################################################################

# Request a proxy object representing the robot to control.
robot = Robot()
name = robot.getName()

# Enable computer keyboard input for user control.
keyboard = Keyboard()
keyboard.enable(EVENT_LOOP_DT)

# Fetch handles for the joint sensors.
j1 = robot.getDevice('joint1')
j2 = robot.getDevice('joint2')
j3 = robot.getDevice('joint3')

# Specify the sampling rate for the joint sensors.
j1.enable(EVENT_LOOP_DT)
j2.enable(EVENT_LOOP_DT)
j3.enable(EVENT_LOOP_DT)

# Fetch handle for the 'base' joint motor.  In this
# example the motor will be controlled as a torque
# motor, bypassing the Webots PID control.
motor1 = robot.getDevice('motor1')
motor1.setTorque(0.0)

motor2 = robot.getDevice('motor2')
motor2.setTorque(0.0)

motor3 = robot.getDevice('motor3')
motor3.setTorque(0.0)

# set properties in RotationalMotor nodes:
motor1.setAvailableTorque(8)
motor2.setAvailableTorque(4)
motor3.setAvailableTorque(4)

# Specify proportional and derivative (damping) gains
# for the position controllers in the main event loop.
P_gain1 = 12.0     # units are N-m / radian
D_gain1 = 0.2     # units are N-m / (radian/sec)

P_gain2 = 9.0     # units are N-m / radian
D_gain2 = 0.1     # units are N-m / (radian/sec)

P_gain3 = 9.0     # units are N-m / radian
D_gain3 = 0.1     # units are N-m / (radian/sec)

# Soft torque limits to avoid error messages.
max_tau1 = 8.0  # units are N-m
max_tau2 = 4.0  # units are N-m
max_tau3 = 4.0  # units are N-m

# Initial position targets.
q_d1 = None
q_d2 = None
q_d3 = None

# The controller estimates velocities using finite
# differencing of the position sensors; these variables
# hold the previous state.
last_q1 = 0
last_q2 = 0
last_q3 = 0


starttime = 0
mode = 0

################################################################
# Define reference poses mapped to keyboard keys.

poses = {
    'r' : (0.75, 1.0, -1.0),          # elbow to machine right (-Y)
    'l' : (0.85, 0.85, -1.0),          # elbow to machine left (+Y)
    'q' : (0.85, 0.85, -1.0), 
    'w' : (0, 0, 0), 
    'o' : (0.85, 0.85, -1.0), 
    'p' : (1.85, 0.85, -1.0)
}

# Set joint position targets to a reference pose indicated by a single character string.
def set_pose(key1, key2, phase):
    global q_d1, q_d2, q_d3
    pose1 = poses.get(key1)
    pose2 = poses.get(key2)
    q_d1 = pose1[0] + (pose2[0] - pose1[0])*phase
    q_d2 = pose1[1] + (pose2[1] - pose1[1])*phase
    q_d3 = pose1[2] + (pose2[2] - pose1[2])*phase

################################################################
# Metronomic choreography for automatic machine.
def set_automatic_targets(t):
    overphase = math.fmod(t, 2.0) / 2.0
    phase = math.fmod(t, 1.0)
    if overphase > 0.5:
        set_pose('l', 'r', phase)
    else:
        set_pose('r', 'l', phase)

################################################################

# Run an event loop until the simulation quits,
# indicated by the step function returning -1.
while robot.step(EVENT_LOOP_DT) != -1:

    # Read simulator clock time.
    t = robot.getTime()

    # Read the new joint positions.
    q1 = j1.getValue()
    q2 = j2.getValue()
    q3 = j3.getValue()

    # Estimate current velocities in radians/sec using finite differences.
    qd1 = (q1 - last_q1) / (0.001 * EVENT_LOOP_DT)
    qd2 = (q2 - last_q2) / (0.001 * EVENT_LOOP_DT)
    qd3 = (q3 - last_q3) / (0.001 * EVENT_LOOP_DT)
    last_q1 = q1
    last_q2 = q2
    last_q3 = q3

    # Read any computer keyboard keypresses.  Returns -1 or an integer keycode while a key is held down.
    key = keyboard.getKey()
    if key != -1 and chr(key).lower() == ' ':
        mode = 1
    if mode == 1:
        starttime = t
        mode = 2
    elif mode == 2:
        phase = math.fmod((t-starttime), 0.1) / 0.1
        set_pose('q', 'w', phase)
        if t-starttime > 0.1:
            mode = 3
    elif mode == 3:
        set_pose('w', 'w', 1)
        if t-starttime > 0.75:
            starttime = t
            mode = 4
    elif mode == 4:
        phase = math.fmod((t-starttime), 1.5) / 1.5
        set_pose('w', 'q', phase)
        if t-starttime > 1.5:
            mode = 5
    elif mode == 5:
        overphase = math.fmod(t-starttime, 1.5) / 1.5
        phase = math.fmod((t-starttime), 0.75) / 0.75
        if overphase > 0.5:
            set_pose('p', 'o', phase)
        else:
            set_pose('o', 'p', phase)
    else:
        set_automatic_targets(t)

    # Calculate new motor torques, limit them, and apply them to the system.
    if q_d1 is None:
        tau1 = 0.0
    else:
        tau1 = P_gain1 * (q_d1 - q1) - D_gain1 * qd1
        tau1 = min(max(tau1, -max_tau1), max_tau1)

    if q_d2 is None:
        tau2 = 0.0
    else:
        tau2 = P_gain2 * (q_d2 - q2) - D_gain2 * qd2
        tau2 = min(max(tau2, -max_tau2), max_tau2)

    if q_d3 is None:
        tau3 = 0.0
    else:
        tau3 = P_gain3 * (q_d3 - q3) - D_gain3 * qd3
        tau3 = min(max(tau3, -max_tau3), max_tau3)

    motor1.setTorque(tau1)
    motor2.setTorque(tau2)
    motor3.setTorque(tau3)
